#include <Wire.h>
#include <ds3231.h>
 
#include<LiquidCrystal.h>

constexpr int BUTTON_LEFT_PIN = 8;
constexpr int BUTTON_CENTER_PIN = 7;
constexpr int BUTTON_RIGHT_PIN = 6;

constexpr int BUZZER_PIN = 9;

constexpr int GLOBAL_DELAY = 150;


struct Time {
  unsigned hour;
  unsigned min;

  bool valid() {
    return hour <= 23 && min <= 59;
  }
};

struct Date {
  unsigned mday;
  unsigned month;
  unsigned year;

  bool valid() {
    if (mday == 0)
      return false;

    if (month == 0 || month > 12)
      return false;
    
    if (month == 1 || month == 3 || month == 5 || month == 7 || month == 8 || month == 10 || month == 12)
      return mday <= 31;
    
    if (month == 4 || month == 6 || month == 9 || month == 11)
      return mday <= 30;
    
    if (month == 2) {
      if (year % 4) return mday <= 28;

      return mday <= 29;
    }

    return false;
  }
};

constexpr int ALIGN_RIGHT = 1;
constexpr int ALIGN_LEFT = 2;
constexpr int ALIGN_CENTER = 3;

class DateFormat {
public:
  virtual String convert(Date date) = 0;

  virtual String presentation() = 0;
};

class TimeFormat {
public:
  virtual String convert(Time time) = 0;

  virtual String presentation() = 0;
};

class FormatContainer {
public:
  DateFormat **dateFormats;
  unsigned dateFormatsSize;
  TimeFormat **timeFormats;
  unsigned timeFormatsSize;

  DateFormat *dateFormat;
  TimeFormat *timeFormat;

  FormatContainer();
};

FormatContainer *formatContainer = new FormatContainer();

class Button {
private:
  int pin;

  long time = 0;
public:
  Button(int pin) : pin(pin) {}

  inline void init() {
    pinMode(pin, INPUT_PULLUP);
  }

  inline bool read() {
    if (abs(millis() - time) < GLOBAL_DELAY)
      return false;
    
    time = millis();

    return digitalRead(pin);
  }
};


class Controls {
public:
  Button *left = new Button(BUTTON_LEFT_PIN);
  Button *right = new Button(BUTTON_RIGHT_PIN);
  Button *center = new Button(BUTTON_CENTER_PIN);
  
  inline void init() {
    left->init();
    right->init();
    center->init();
  }
};

Controls *controls = new Controls();


class Song {
public:
  int *melody;
  int length;
  int tempo;
  int wholenote;
  String name;

  inline Song(int *melody, int length, int tempo, String name)
      : melody(melody), length(length), tempo(tempo), name(name) {
      wholenote = (60000 * 4) / tempo;

  }
};

class SongCollection {
public:
  int size;
  Song **collection;
  Song *currentSong;

  SongCollection();
};

SongCollection *songCollection = new SongCollection();

class MusicPlayer {
private:
  int noteIndex;
  long timee = 0;

  inline void playNote() {
    int divider = songCollection->currentSong->melody[noteIndex + 1];
    int noteDuration = 0;
    if (divider > 0) {
      noteDuration = songCollection->currentSong->wholenote / divider;
    } else if (divider < 0) {
      noteDuration = songCollection->currentSong->wholenote / abs(divider);
      noteDuration *= 1.5;
    }

    if (noteDuration < millis() - timee) {
      tone(BUZZER_PIN, songCollection->currentSong->melody[noteIndex], noteDuration * 0.9);
      timee = millis();

      noteIndex += 2;
      if (noteIndex >= songCollection->currentSong->length)
          noteIndex = 0;
        
    }
  }
public:
  inline int play() {
    noteIndex = 0;
    while (1) {

      if (controls->left->read()) {
          noTone(BUZZER_PIN);
          return -1;
      }
      
      if (controls->center->read()) {
          noTone(BUZZER_PIN);
          return 0;
      }

      if (controls->right->read()) {
          noTone(BUZZER_PIN);
          return 1;
      }
      
      playNote();
    }
  }

};

MusicPlayer *musicPlayer = new MusicPlayer();


class Screen {
private:
  int WIDTH = 16;
  int HEIGHT = 2;
  int DELAY_TIME = 100;

  LiquidCrystal lcd = LiquidCrystal(12, 11, 5, 4, 3, 2);

  String lastText = "";
  String lastPrefix = "";

  String addPrefix(String text, int align) {
    if (align == ALIGN_LEFT)
        return text;

    if (text.length() >= WIDTH)
        return text;

    int step = align == ALIGN_RIGHT ? 1 : 2;
    String prefix = "";
    for (int i = 0; i < WIDTH - text.length(); i += step) {
        prefix += " ";
    }

    lastPrefix = prefix;
    lastText = text;

    return prefix + text;
  }

public:
  Screen() {}

  void init() {
    lcd.begin(WIDTH, HEIGHT);
  }

  void display(String text, int align) {
    lcd.clear();
    text = addPrefix(text, align);
    
    lcd.print(text);
  }

  void display(String text) {
    display(text, ALIGN_CENTER);
  }

  void change(int startIndex, String text) {
    startIndex += lastPrefix.length();

    lcd.setCursor(startIndex % WIDTH, startIndex / WIDTH);
    lcd.print(text);
    delay(10);
  }

  void displayLines(String first, String second) {
    lcd.clear();

    first = addPrefix(first, ALIGN_CENTER);
    second = addPrefix(second, ALIGN_CENTER);

    lcd.print(first);
    delay(10);
    lcd.setCursor(0, 1);
    lcd.print(second);
    delay(10);
  }

  void displayFull(String text) {
    Serial.println(text);
    lcd.clear();

    lcd.print(text.substring(0, WIDTH));
    lcd.setCursor(0, 1);
    lcd.print(text.substring(WIDTH));
    delay(10);
  }
};

Screen *screen = new Screen();

class Clock {
private:
  struct ts t;  

  unsigned long delay = 0;

public:
  void get() {
    DS3231_get(&t);
    if (delay + 1000 < millis()) {
      DS3231_set(t);
      delay = millis();
    }

    Serial.println(String(t.hour) + " " + String(t.min) + " " + t.sec);

  }

  Time getTime() {
    get();

    return {
      .hour = t.hour,
      .min = t.min
    };
  }

  Date getDate() {
    get();

    return {
      .mday = t.mday,
      .month = t.mon,
      .year = t.year
    };
  }

  void setTime(Time time) {
    DS3231_get(&t);

    t.hour = time.hour;
    t.min = time.min;
    t.sec = 0;

    DS3231_set(t);
    delay = millis();
  }

  void setDate(Date date) {
    DS3231_get(&t);

    t.mday = date.mday;
    t.mon = date.month;
    t.year = date.year;

    DS3231_set(t);
    delay = millis();
  }
};

Clock *clock = new Clock();

class Alarm {
public:
  Time time = { 100, 100 };

  bool execute() {
    if (!time.valid()) return false;

    Time current = clock->getTime();
    if (time.hour != current.hour || time.min != current.min) return false;

    screen->display("ALARMA");
    musicPlayer->play();

    time.hour = 100;
    time.min = 100;

    return true;
  }
};

Alarm *alarm = new Alarm();

class TimePicker {
private:
  char _pickValue(int index) {
    String values = "0123456789";
    int valueIndex = 0;
    bool changed = true;


    while (1) {
      if (controls->center->read()) {
        screen->change(index, String(values[valueIndex])); 

        return values[valueIndex];
      }

      if (controls->left->read()) {
        valueIndex = max(0, valueIndex - 1);
        changed = true;
      }

      if (controls->right->read()) {
        valueIndex = min(values.length(), valueIndex + 1);
        changed = true;
      }

      if (changed) {
        screen->change(index, String(values[valueIndex]));
        changed = false;
      }
    }
  }
  Time _pickTime() {
    int current = 0;
    bool showIndex = true;
    String time = "__:__";

    screen->display(time);

    for (int index = 0; index < 5; index++) {
      if (index != 2)
        time[index] = _pickValue(index);
    }


    return {
      .hour = time.substring(0, 2).toInt(),
      .min = time.substring(3, 5).toInt(),
    };
  }

  Date _pickDate() {
    int current = 0;
    bool showIndex = true;
    String date = "__/__/____";

    screen->display(date);

    for (int index = 0; index < 10; index++) {
      if (index != 2 && index != 5)
        date[index] = _pickValue(index);
    }

    return {
      .mday = date.substring(0, 2).toInt(),
      .month = date.substring(3, 5).toInt(),
      .year = date.substring(6, 10).toInt()
    };
  }
public:
  Time pickTime() {

    while (1) {
      Time value = _pickTime();
      
      if (value.valid())
        return value;
      
      screen->display("INVALID");
      delay(200);
    }

  }
  
  Date pickDate() {
    while (1) {
      Date value = _pickDate();
      
      if (value.valid())
        return value;
      
      screen->display("INVALID");
      delay(200);
    }
  }
};

TimePicker *timePicker = new TimePicker();

class MenuCommand {
public:
    virtual String getName() = 0;

    virtual void execute() = 0;
};

class Menu {
public:
  MenuCommand **commands;
  unsigned commandsSize;

  void initCommands();
  Menu() {
    initCommands();
  }

  void execute() {
    unsigned index = 0;
    bool changed = true;

    while (1) {
      if (controls->left->read()) {
        changed = true;
        if (index == 0) {
          index = commandsSize - 1;
        }
        else 
          index -= 1;
      }

      if (controls->right->read()) {
        changed = true;
        index = (index + 1) % commandsSize;
      }

      if (controls->center->read()) {
        commands[index]->execute();
        break;
      }

      if (changed) {
        screen->displayFull(commands[index]->getName());
        changed = false;
      }
    }
  }
};

Menu *menu = new Menu();


class ShowTime {
private:
    String prevDate = "";
    String prevTime = "";
public:
    void show(bool clean = false) {
        if (clean) {
            prevDate = "";
            prevTime = "";
        }

        bool changed = false;
        Time time = clock->getTime();
        Date date = clock->getDate();

        String showTime = formatContainer->timeFormat->convert(time);
        String showDate = formatContainer->dateFormat->convert(date);

        if (prevTime != showTime) {
            changed = true;
            prevTime = showTime;
        }

        if (prevDate != showDate) {
            changed = true;
            prevDate = showDate;
        }

        if (changed) {
            screen->displayLines(showTime, showDate);
            changed = false;
        }

        delay(10);
    }
};

ShowTime *showTime = new ShowTime();

class Project {
public:

    void init() {
        controls->init();
        screen->init();
    }

    void execute() {
        showTime->show();
        alarm->execute();

        if (controls->center->read()) {
            menu->execute();
            showTime->show(true);
        }
        delay(100);

        // musicPlayer->play();
        // delay(1000);
    }
};


Project *project = new Project();


void setup() {
  Wire.begin();
  DS3231_init(DS3231_CONTROL_INTCN);
  project->init();
  pinMode(BUZZER_PIN, OUTPUT);
  Serial.begin(9600);
}

void execute() {
    project->execute();
}

void loop() {
    execute();
}



// Formats
class NormalTimeFormat : public TimeFormat {
public:
  inline String convert(Time time) {
    char buf[19];
    sprintf(buf, "%02u:%02u", time.hour, time.min);

    return buf;
  }

  inline String presentation() override {
    return "HH:MM";
  }
};


class APTimeFormat : public TimeFormat {
public:
  inline String convert(Time time) {
    unsigned minutes = time.min;
    unsigned hours = time.hour;

    const char *dayPart = hours / 12 ? "PM" : "AM";
    hours %= 12;
    if (hours == 0) {
        hours = 12;
    }
    char buffer[30];
    sprintf(buffer, "%u:%02u %s", hours, minutes, dayPart);

    return buffer;
  }

  inline String presentation() override {
    return "HH:MM AM";
  }
};

class DateFormat1 : public DateFormat {
public:
  inline String convert(Date date) {
      char buf[30];
      sprintf(buf, "%02u/%02u/%04u", date.mday, date.month, date.year);

      return buf;
  }
  
  inline String presentation() {
      return "15/10/2020";
  }
};

class DateFormat2 : public DateFormat {
private:
    String *months;
public:

    DateFormat2() {
        months = new String[12];
        months[0] = "Ian";
        months[1] = "Feb";
        months[2] = "Mar";
        months[3] = "Apr";
        months[4] = "Mai";
        months[5] = "Iun";
        months[6] = "Iul";
        months[7] = "Aug";
        months[8] = "Sep";
        months[9] = "Oct";
        months[10] = "Noi";
        months[11] = "Dec";
    }

    inline String convert(Date date) {
        unsigned day = date.mday;
        unsigned month = date.month;
        unsigned year = date.year;

        return String(day) + " " + months[month - 1] + " " + String(year);
    }
    
    inline String presentation() {
        return "15 Oct 2020";
    }
};

class DateFormat3 : public DateFormat {
private:
    String *months;
public:

    DateFormat3() {
        months = new String[12];
        months[0]  = "Ianuarie";
        months[1]  = "Februarie";
        months[2]  = "Martie";
        months[3]  = "Aprilie";
        months[4]  = "Mai";
        months[5]  = "Iunie";
        months[6]  = "Iulie";
        months[7]  = "August";
        months[8]  = "Septembrie";
        months[9]  = "Octombrie";
        months[10] = "Noiembrie";
        months[11] = "Decembrie";
    }

    inline String convert(Date date) {
        unsigned day = date.mday;
        unsigned month = date.month;

        return String(day) + " " + months[month - 1];
    }
    
    inline String presentation() {
        return "15 Octombrie";
    }
};

FormatContainer::FormatContainer() {
    dateFormatsSize = 2;
    dateFormats = new DateFormat*[dateFormatsSize];

    dateFormats[0] = new DateFormat1();
    dateFormats[1] = new DateFormat2();

    timeFormatsSize = 2;
    timeFormats = new TimeFormat*[timeFormatsSize];

    timeFormats[0] = new NormalTimeFormat();
    timeFormats[1] = new APTimeFormat();

    dateFormat = dateFormats[1];
    timeFormat = timeFormats[0];
}

class SetAlarmCommand : public MenuCommand {
public:
    String getName() {
        return alarm->time.valid() ? "Opreste alarma" : "Seteaza alarma";
    }

    void execute() {
        if (alarm->time.valid()) {
            alarm->time = {100, 100};
            return;
        }

        alarm->time = timePicker->pickTime();
        delay(10);
    }
};

class SetHourCommand : public MenuCommand {
public:
  String getName() {
    return "Seteaza ora";
  }

  void execute() {
    Time time = timePicker->pickTime();
    clock->setTime(time);
    delay(10);
  }
};


class SelectHourFormatCommand : public MenuCommand {
public:
  String getName() {
    return "Seteaza formatul orei";
  }

  void execute() {
    unsigned index = 0;
    bool changed = true;

    while (1) {
      if (controls->left->read()) {
        if (index == 0) {
          index = formatContainer->timeFormatsSize - 1;
        }
        else {
          index = (index - 1) % formatContainer->timeFormatsSize;
        }
        changed = true;
      }

      if (controls->right->read()) {
        index = (index + 1) % formatContainer->timeFormatsSize;
        changed = true;
      }

      if (controls->center->read()) {
        formatContainer->timeFormat = formatContainer->timeFormats[index];
        break;
      }

      if (changed) {
        screen->display(formatContainer->timeFormats[index]->presentation());
        changed = false;
      }
    }
  }
};

class SetDateCommand : public MenuCommand {
public:
  String getName() {
    return "Seteaza data";
  }

  void execute() {
    Date date = timePicker->pickDate();
    clock->setDate(date);
  }
};

class SelectDateFormatCommand : public MenuCommand {
public:
  String getName() {
      return "Seteaza formatul datii";
  }

  void execute() {
    unsigned index = 0;
    bool changed = true;

    while (1) {
      if (controls->left->read()) {
        if (index == 0) {
          index = formatContainer->dateFormatsSize - 1;
        } else {
          index = (index - 1) % formatContainer->dateFormatsSize;
        }
        changed = true;
      }

      if (controls->right->read()) {
        index = (index + 1) % formatContainer->dateFormatsSize;
        changed = true;
      }

      if (controls->center->read()) {
        formatContainer->dateFormat = formatContainer->dateFormats[index];
        break;
      }

      if (changed) {
        screen->display(formatContainer->dateFormats[index]->presentation());
        changed = false;
      }
    }
  }
};

class SelectAudioCommand : public MenuCommand {
public:
    String getName() {
        return "Seteaza melodia";
    }

    void execute() {
        unsigned index = 0;
        int x = 1;
        while (x) {
            screen->display(songCollection->collection[index]->name);

            x = musicPlayer->play();

            index += x;
            index %= songCollection->size;

            songCollection->currentSong = songCollection->collection[index];
        }
    }
};

class ExitCommand : public MenuCommand {
    String getName() {
        return "Iesire";
    }

    void execute() {}
};

void Menu::initCommands() {
    commandsSize = 7;
    commands = new MenuCommand*[commandsSize];

    commands[0] = new SetAlarmCommand();
    commands[1] = new SetHourCommand();
    commands[2] = new SelectHourFormatCommand();
    commands[3] = new SetDateCommand();
    commands[4] = new SelectDateFormatCommand();
    commands[5] = new SelectAudioCommand();
    commands[6] = new ExitCommand();
}

// MELODII


#define NOTE_B0  31
#define NOTE_C1  33
#define NOTE_CS1 35
#define NOTE_D1  37
#define NOTE_DS1 39
#define NOTE_E1  41
#define NOTE_F1  44
#define NOTE_FS1 46
#define NOTE_G1  49
#define NOTE_GS1 52
#define NOTE_A1  55
#define NOTE_AS1 58
#define NOTE_B1  62
#define NOTE_C2  65
#define NOTE_CS2 69
#define NOTE_D2  73
#define NOTE_DS2 78
#define NOTE_E2  82
#define NOTE_F2  87
#define NOTE_FS2 93
#define NOTE_G2  98
#define NOTE_GS2 104
#define NOTE_A2  110
#define NOTE_AS2 117
#define NOTE_B2  123
#define NOTE_C3  131
#define NOTE_CS3 139
#define NOTE_D3  147
#define NOTE_DS3 156
#define NOTE_E3  165
#define NOTE_F3  175
#define NOTE_FS3 185
#define NOTE_G3  196
#define NOTE_GS3 208
#define NOTE_A3  220
#define NOTE_AS3 233
#define NOTE_B3  247
#define NOTE_C4  262
#define NOTE_CS4 277
#define NOTE_D4  294
#define NOTE_DS4 311
#define NOTE_E4  330
#define NOTE_F4  349
#define NOTE_FS4 370
#define NOTE_G4  392
#define NOTE_GS4 415
#define NOTE_A4  440
#define NOTE_AS4 466
#define NOTE_B4  494
#define NOTE_C5  523
#define NOTE_CS5 554
#define NOTE_D5  587
#define NOTE_DS5 622
#define NOTE_E5  659
#define NOTE_F5  698
#define NOTE_FS5 740
#define NOTE_G5  784
#define NOTE_GS5 831
#define NOTE_A5  880
#define NOTE_AS5 932
#define NOTE_B5  988
#define NOTE_C6  1047
#define NOTE_CS6 1109
#define NOTE_D6  1175
#define NOTE_DS6 1245
#define NOTE_E6  1319
#define NOTE_F6  1397
#define NOTE_FS6 1480
#define NOTE_G6  1568
#define NOTE_GS6 1661
#define NOTE_A6  1760
#define NOTE_AS6 1865
#define NOTE_B6  1976
#define NOTE_C7  2093
#define NOTE_CS7 2217
#define NOTE_D7  2349
#define NOTE_DS7 2489
#define NOTE_E7  2637
#define NOTE_F7  2794
#define NOTE_FS7 2960
#define NOTE_G7  3136
#define NOTE_GS7 3322
#define NOTE_A7  3520
#define NOTE_AS7 3729
#define NOTE_B7  3951
#define NOTE_C8  4186
#define NOTE_CS8 4435
#define NOTE_D8  4699
#define NOTE_DS8 4978
#define REST      0


int asabranca[] = {
    NOTE_G4,8, NOTE_A4,8, NOTE_B4,4, NOTE_D5,4, NOTE_D5,4, NOTE_B4,4, 
    NOTE_C5,4, NOTE_C5,2, NOTE_G4,8, NOTE_A4,8,
};

int bloodytears[] = {

  REST, 4, NOTE_G5, 4,
  NOTE_A5, 4, NOTE_AS5, 4,
  NOTE_A5, 4, NOTE_F5, 4,
  NOTE_A5, 4, NOTE_G5, 4,
  REST, 4,  
};

SongCollection::SongCollection() {
    size = 2;
    collection = new Song*[size];
    collection[0] = new Song(asabranca, sizeof(asabranca) / sizeof(int) / 2, 120, "Asabranca");
    collection[1] = new Song(bloodytears, sizeof(bloodytears) / sizeof(int) / 2, 144, "Bloodytears");
    currentSong = collection[0];
}